ndsCreateDevice "TechnosoftTML", "", "FILE=/tmp/,NAXIS=4,DEV_PATH=/dev/ttyS0,HOST_ID=1,
    AXIS_SETUP_0=$(SUPPORT)/motorTechnosoft/tml_lib/config/config.txt,AXIS_ID_0=1,AXIS_HOMING_SW_0=LSN,
    AXIS_SETUP_1=$(SUPPORT)/motorTechnosoft/tml_lib/config/config.txt,AXIS_ID_1=2,AXIS_HOMING_SW_1=LSN,
    AXIS_SETUP_2=$(SUPPORT)/motorTechnosoft/tml_lib/config/config.txt,AXIS_ID_2=3,AXIS_HOMING_SW_2=LSN,
    AXIS_SETUP_3=$(SUPPORT)/motorTechnosoft/tml_lib/config/config.txt,AXIS_ID_3=4,AXIS_HOMING_SW_3=LSN"

